package JQJoiner;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.math.*;
import java.net.*;
import java.util.*;
import JQJoiner.GIFDecoderPkg.*;
import JQJoiner.GIFEncoderPkg.*;

public class CMReplacer implements Messenger
{

  private String cmString_ = null;
  private String pixString_ = null;
  private String outputString_ = null;
  private FractalSpec args_ = null; // Stores the current FractalSpec.
  private RGBxSizeSequence baseColorMap_ = null;
  private RGBxSizeSequence pixColorMap_ = null;
  
  private boolean abort_ = false;
  private GIFEncoder encoder_ = null;
  private OutputStream outStream_ = null;

  private byte [] pixels_ = null;
  private int [] translation_ = null;
  private int width_ = 0;
  private int height_ = 0;
  private int transparentPixelIndex_ = -1; // -1 means none to the GIFEncoder.

  
  CMReplacer (String cmPath, String pixPath, String outPath)
  {
    cmString_ = cmPath;
    pixString_ = pixPath;
    outputString_ = outPath;
  }
  
  
  public void replaceColorMap ()
    // Top level action method.
  {
    System.out.println(" ");
    System.out.println("Replacing Color Map of " +  pixString_ + " with that of " + cmString_);    
    
    // Do the input.
    try {
      baseColorMap_ = getColorMap(cmString_);
      pixColorMap_ = loadArgsThenGIF(pixString_); // Sets args_ and pixels_, etc.
      // fixPixelZero(); - Now that my encoder is right this is not needed.
      translation_ = findTranslation(baseColorMap_, pixColorMap_);
      translateAndOutput();
    } catch (Throwable e) {
      System.out.println("CMReppacer.getColorMap: catch " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()));
      System.out.println("CMReppacer ABORTING");
    }
  }
  
  public void message(String arg)
  {
    System.out.println(arg);
  }
  
  private RGBxSizeSequence getColorMap (String pathname) throws Throwable
  {
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    Throwable exception1 = null;
    System.out.println(" ");
    System.out.println("Getting color map from " +  pathname);    
    // Here we try to open the stream and extract the arguments only.
    try {
      stream1 = toInputStream(pathname, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);
      gifDecoder = new GIFDecoder(stream1);
      String argsString = gifDecoder.getApplicationData("JavaQuat", "020");
      if (argsString != null) {
//         StringBufferInputStream argsStream = new StringBufferInputStream(argsString);
	StringReader argsStream = new StringReader(argsString);
        JQArgReader reader = new JQArgReader();
        // parent.exitImageSaver gets passed argsReadFromFile_ on exitButton.
        args_ = new FractalSpec(reader.readArgFile(argsStream, -1, this));
      }
      if (gifDecoder.transparentPixelP()) {
        transparentPixelIndex_ = gifDecoder.transparentPixelIndex();
        if (transparentPixelIndex_ < 0) {
          transparentPixelIndex_ += 256;
        }
      }
      return gifDecoder.globalRGBTable();
    } catch (Throwable e) {
      System.out.println("CMReppacer.getColorMap: catch " + e.toString() + " " +
                         (e.getMessage() == null ? "" : e.getMessage()) );
      e.printStackTrace();
      exception1 = e;
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }
    System.out.println("CMOeplacer.getColorMap about to exit with" + 
                       (exception1 == null ? "out " : " ") + "error.");
    if (exception1 != null)
      throw exception1;
    return null;
  }
  
  
  
  public RGBxSizeSequence loadArgsThenGIF (String pathname) throws Throwable
  {
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    Throwable exception1 = null;
    System.out.println(" ");
    System.out.println("Opening file " +  pathname);    
    // Here we try to open the stream and extract the arguments only.
    try {
      stream1 = toInputStream(pathname, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);
      gifDecoder = new GIFDecoder(stream1);
      String argsString = gifDecoder.getApplicationData("JavaQuat", "014");
      if (argsString != null) {
//         StringBufferInputStream argsStream = new StringBufferInputStream(argsString);
	StringReader argsStream = new StringReader(argsString);
      
        JQArgReader reader = new JQArgReader();
        // parent.exitImageSaver gets passed argsReadFromFile_ on exitButton.
        args_ = new FractalSpec(reader.readArgFile(argsStream, -1, this));
      }
      width_ = gifDecoder.width();
      height_ = gifDecoder.height();
      pixels_ = gifDecoder.pixelsAsIndices();
      if (pixels_ == null) {
        System.out.println("CMReplacer.loadArgsThenGIF: Empty pixels_");
        throw new IOException("Empty pixels_ in loadArgsThenGIF");
      }
      
    } catch (Throwable e) {
      System.out.println("CMReplacer.loadArgsThenGIF: catch " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()));
      e.printStackTrace();
      exception1 = e;
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }  
    
    if (exception1 != null)
      throw exception1;
    return gifDecoder.globalRGBTable();
  }
  
  
  private void fixPixelZero ()
    // In a gif image the first pixel index must be zero.  In order to save the whole
    // color map in my images (except position 0) I set the 0th element of the color map
    // to the value of the zeroth pixel.  Since in joining images we use the color map of
    // the first image, subsequent images need to have their zeroth pixel value restored.
  {
    byte [] rgb = pixColorMap_.byteTable;
    for (int i = 3; i < rgb.length; i += 3) {
      if (rgb[0] == rgb[i] &&      //reds match
          rgb[1] == rgb[i + 1] &&  // greens match
          rgb[2] == rgb[i + 2]) {
        pixels_[0] = (byte)(i / 3);
        return;
      }
    }
    System.out.println("CMReplacer fixPixelZero failed.");
  }
  
  
  public InputStream toInputStream (String pathname1, boolean testOnly_p)
  {
     InputStream stream1 = null;
     try {
       System.out.println("VertJoiner opening " + pathname1);    
       File testExistance = new File(pathname1);
       if (!testExistance.exists()) {
         throw new IOException("File DOES NOT EXIST - no read done.");
//       } else if (!testExistance.canWrite()) {
//         throw new IOException("File cannot be written to.");
       }
       if (testOnly_p)
         stream1 = null;
       else
         stream1 = new FileInputStream(pathname1);
     } catch (Throwable e) {
       System.out.println("CMReplacer FAILED TO OPEN " + pathname1);    
       stream1 = null;
       abort_ = true;
     } finally {
       return stream1;
     }
  }
  
  public OutputStream toOutputStream (String pathname1, boolean testOnly_p)
  {
     OutputStream stream1 = null;
     try {
  System.out.println("Opening output file " + pathname1);    
       File testExistance = new File(pathname1);
       if (testExistance.exists()) {
         throw new IOException("Already Exists or invalid - no write done.");
       }
       if (testOnly_p)
         stream1 = null;
       else
         stream1 = new FileOutputStream(pathname1);
     } catch (Throwable e) {
       System.out.println("CMReplacer.toOutputStream " + 
                          (e.getMessage() == null ? e.toString() : e.getMessage()));
       stream1 = null;
     } finally {
       return stream1;
     }
  }
  
  private int [] findTranslation (RGBxSizeSequence baseSeq, RGBxSizeSequence pixSeq)
  {
    int [] answer = new int[256];
    byte [] baseRGB = baseSeq.byteTable;
    byte [] pixRGB = pixSeq.byteTable;
    boolean success = false;
    int bestDiff;
    int bestIndex = 0;
    for(int p = 0; p < pixRGB.length; p += 3) {
      bestDiff = 1 << 20;
      success = false;
      for(int b = 0; b < baseRGB.length; b += 3) { //b=3 to skip bogus 0th color map entry.
        int rDiff = baseRGB[b] - pixRGB[p];
        int gDiff = baseRGB[b+1] - pixRGB[p+1];
        int bDiff = baseRGB[b+2] - pixRGB[p+2];
        int diff = (rDiff * rDiff) + (gDiff * gDiff) + (bDiff * bDiff);
        if (diff == 0) {
          bestIndex = b/3;
          break;
        }
        if (diff < bestDiff) {
          bestDiff = diff;
          bestIndex = b/3;
        }
      }
      answer[p/3] = bestIndex;
    }
    return answer;
  }
  
  private void translateAndOutput ()
  {
    // Translate the pixels
    for (int i = 0; i < pixels_.length; i++) {
      int index = pixels_[i];
      if (index < 0)
        index += 256;
      int newValue = translation_[index];
      if (newValue > 127)
        newValue -= 256;
      pixels_[i] = (byte)newValue;
    }
    
    // OUTPUT
    beginOutput();
    copyToOutput();
    finishOutput();
  }

  private void beginOutput()
  {   
    // Set up the output spec. - This needs work.    
    System.out.println("Begin writing preparation.");
      
    // Write out the images.
    try {
      outStream_ = toOutputStream(outputString_, false);
      encoder_ = new GIFEncoder(width_, height_, baseColorMap_);
      String arglist = (args_ == null ? null : args_.toString(1, " Color Map Replace"));
      encoder_.javaQuatWriteHeader(outStream_, arglist, false, transparentPixelIndex_);
    } catch (Throwable e) {
      System.err.println("CMReplacer.beginOutput error: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
    }
  }
  
  
  private void copyToOutput()
  {
//  System.out.println("  copyToOutput entry.");
    try {
      encoder_.javaQuatWritePixels(pixels_, 0, width_ * height_);
    } catch (Throwable e) {
      System.err.println("copyToOutput error: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
    }
//  System.out.println("copyToOutput exit.");
  }
  
  
  private void finishOutput ()
  {
//  System.out.println("  finishOutput entry.");
    try {
      encoder_.javaQuatWriteFooter();
      System.err.println("Color Map Replacement IS COMPLETE.");
    } catch (Throwable e) {
      System.err.println("finishOutput error: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
    
    }finally {
      if (outStream_ != null)
        try {
          outStream_.close(); // We didn't close it above.
        } catch (Throwable e) {}
    } 
  }      

}
