package JQJoiner;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.math.*;
import java.net.*;
import java.util.*;
import JQJoiner.GIFDecoderPkg.*;
import JQJoiner.GIFEncoderPkg.*;

public class ChunckJoiner extends Canvas implements Runnable
{
  static int count_ = 0;

  private String inputString_ = null;
  private String outputStringBase_ = null;
  private String outputString_ = null;
  private Vector panelStrings_;
  int vChunk_;
  int hChunk_; 
  int vFirst_overall_;
  int vLast_overall_;
  int hFirst_overall_;
  int hLast_overall_;
  int vFirst_;
  int vLast_;
  int hFirst_;
  int hLast_;
  
  private Messenger messenger_ = null;
  private FractalSpec args_ = null; // Stores the zeroth FractalSpec.
  private RGBxSizeSequence color_map_ = null;
  private int row_count_ = 0;
  private int col_count_ = 0;

  private int vIndex_;
  private int hIndex_;
  private int totalWidth_ = 0;
  private int row_width_ = 0; // Computer for each row if component images & checked 
                              // against totalWidth_.
  private int panel_height_ = 0;

  private boolean abort_ = false;
  private GIFEncoder encoder_ = null;
  private OutputStream outStream_ = null;
  
  private Object [] pixels_;
  private int [] widths_;
  private int height_ = -1; // Values should all be the same.
  private Thread runner_ = null;  // This is referenced only in unused method join.
  private boolean running_p_ = false;

  private boolean transparentPixelP_ = false;
  private int transparentPixelIndex_ = 0;

  
  ChunckJoiner (String inputString, int v1, int vn, int h1, int hn, int ych, int xch, 
            String outputStringIn)
    // Constructor.
  {
    inputString_ = inputString;
    vChunk_ = ych;
    hChunk_ = xch; 
    vFirst_overall_ = v1;
    vLast_overall_ = vn;
    hFirst_overall_ = h1;
    hLast_overall_ = hn;
    outputStringBase_ = outputStringIn;
  }
  
  
  
  public /* synchronized */ void join (Messenger messenger)
    // This is not currently used, instead we use prep and run.
  {
    messenger_ = messenger;
    abort_ = false;
    
    if (running_p_) {
      // This isn't going to happen, since we always make a new ChunckJoiner at present.
      messenger_.message("** A join is already running.  No action taken.");
      return;
    }
    running_p_ = true;
    runner_ = new Thread(this, "chunk-joiner");
    runner_.setPriority(Thread.MIN_PRIORITY);
    runner_.start();
    
  }
  
  public boolean prep (Messenger messenger)
    // This is called just before run..  Return T if OK to run.
  {
    messenger_ = messenger;
    abort_ = false;
    // runIt();
    
    if (running_p_) {
      // This isn't going to happen, since we always make a new SqJoiner at present.
      messenger_.message("** A join is already running.  No action taken.");
      return false;
    }
    running_p_ = true;
    return true;
  }
  
  
  public void abortJoin ()
  {
    // messenger_.message("Abort join.");
    System.out.println("Abort join.");
    abort_ = true;
  }
  
  public void run ()
    // Top level action method.
  {
    int yChunk = (int)Math.ceil(vFirst_overall_ / vChunk_);
    int firstXChunkNumber =
                 (int)Math.ceil(hFirst_overall_ / hChunk_);
    int xChunk;
    int yMax = (int)Math.ceil((vLast_overall_ - vFirst_overall_ + 1) / vChunk_);
    int xMax = (int)Math.ceil((hLast_overall_ - hFirst_overall_ + 1) / hChunk_);
    pixels_ = new Object [hChunk_];  //Individual arrays added in loadGIF.
    widths_ = new int [hChunk_];    
    for(vFirst_ = vFirst_overall_; vFirst_ <= vLast_overall_; vFirst_ += vChunk_) {
      vLast_ = Math.min(vLast_overall_, vFirst_ + vChunk_ - 1);
      row_count_ = vLast_ - vFirst_ + 1;
      xChunk = firstXChunkNumber;
      for(hFirst_ = hFirst_overall_; hFirst_ <= hLast_overall_; hFirst_ += hChunk_) {
        if (abort_) {
          break;
        }
        hLast_ = Math.min(hLast_overall_, hFirst_ + hChunk_ - 1);
        col_count_ = hLast_ - hFirst_ + 1;
        outputString_ = (outputStringBase_ + "y" + 
                         ZeroPaddedIntString(yChunk, yMax) + "-x" + 
                         ZeroPaddedIntString(xChunk, xMax) + ".gif");
        joinChunk();
        xChunk++;
      }
      yChunk++;
      if (abort_) {
        break;
      }
    }
    running_p_ = false;
    System.gc();
  }
  
  
  void joinChunk ()
  {
    System.out.println(" ");
    System.out.println("Beginning Joining " + outputString_); 
    
    System.out.println("ChunckJoiner.constructor  row_count_ = " + row_count_ + 
                       " Y = " + vFirst_ + " to = " + vLast_ +
                       " X = " + hFirst_ + " to = " + hLast_);
    
    totalWidth_ = 0; // Computed in beginOutput() below.
    PathVector rowPathnames = 
        new PathVector(vFirst_, vLast_, inputString_, "+");
    try {
      for(vIndex_ = 0; vIndex_ < row_count_; vIndex_++) {
        if (abort_) {
          break;
        }
        panelStrings_ = 
          new PathVector(hFirst_, hLast_, 
                         (String)rowPathnames.elementAt(vIndex_), "*");
        row_width_ = 0; // Computed in loadGIF below.
        for(hIndex_ = 0; hIndex_ < col_count_; hIndex_++) {
//          messenger_.clearMessages();
          loadGIF(hIndex_); 
          if (abort_) {
            break;
          }
        }
        if (abort_) { 
          break;
        } else {
          int startTail = outputString_.length() - 30; // up to  30 char.
          if (startTail < 0) {
            startTail = 0;
          }
          messenger_.message("Working on row " + vIndex_ + " of ..." +
                             outputString_.substring(Math.max(0,startTail)) );
          if (vIndex_ == 0) {
            beginOutput();  // Among other things this computes totalWidth_ used below.
          }
          if (row_width_ != totalWidth_) {
            throw new IOException("Component row has wrong width.");
          }
          attemptOutput();
        }
      }
    } catch (Throwable e) {
      messenger_.message("SqJoiner.joinChunk: problem: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
      System.err.println("SqJoiner.joinChunk: problem: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
      e.printStackTrace();
    }
    finishOutput(); // Tries to close file.
  }
  
  
  public void loadGIF (int pixIndex)
  {
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    String pathname = (String)panelStrings_.elementAt(pixIndex);
    System.out.println(" ");
    // Here we try to open the stream and extract the arguments only.
    try {
      stream1 = toInputStream(pathname, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);

      gifDecoder = new GIFDecoder(stream1);
      String argsString = gifDecoder.getApplicationData("JavaQuat", "020");
      args_ = null;
      if (argsString != null) {
//         StringBufferInputStream argsStream = new StringBufferInputStream(argsString);
	StringReader argsStream = new StringReader(argsString);
        JQArgReader reader = new JQArgReader();
        // parent.exitImageSaver gets passed argsReadFromFile_ on exitButton.
        args_ = new FractalSpec(reader.readArgFile(argsStream, -1, messenger_));
      }
      if (args_ == null) {
        System.out.println("ChunkJoiner.loadGIF: Empty spec returned by loadGIF");
      }

      if (pixIndex == 0) {
        panel_height_ = gifDecoder.getHeight();
      } else if (panel_height_ != gifDecoder.getHeight()) {
        throw new IOException("Component picture has wrong height: " + gifDecoder.getHeight() + 
                              ".  Should be " + panel_height_ + " like 1st row component.");
      }
      widths_[pixIndex] = gifDecoder.getWidth();
      row_width_ += widths_[pixIndex];
      System.out.println("  SqJoiner:  which is image " +  pixIndex +
                         " of height " + gifDecoder.getHeight() + " and width " + widths_[pixIndex]);    
      
      pixels_[pixIndex] = gifDecoder.pixelsAsIndices();
      if (pixels_[pixIndex] == null) {
        System.out.println("SqJoiner.loadGIF: Empty pixels_[" +
                           pixIndex + "]");
        abort_ = true;
        throw new IOException("Empty pixels_ in loadGIF");
      }
      
    } catch (Throwable e) {
      //messenger_.message("SqJoiner.loadGIF" +
      //                   (e.getMessage() == null ? e.toString() : e.getMessage()));
      System.out.println("SqJoiner.loadGIF: catch " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()));
      e.printStackTrace();
      abort_ = true;
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }  
    
    if (pixIndex == 0 && vIndex_ == 0) {
      System.out.println("  Preparing color map.");    
      // Set up the byte and int versions of the colormap.
      color_map_ = gifDecoder.globalRGBTable();
      if (color_map_ == null) {
        System.err.println("SqJoiner.loadGIF: Failed to set color_map_.");
        abort_ = true;
      }
      transparentPixelP_ = gifDecoder.transparentPixelP();
      if (transparentPixelP_) {
        transparentPixelIndex_ = gifDecoder.transparentPixelIndex();
      } else {
        transparentPixelIndex_ = -1;
      }
    }
  }
  
 
  private String ZeroPaddedIntString (int valueNow, int max)
  {
    StringBuffer buf = new StringBuffer(8);
    int max_digits = (int)Math.floor(Math.log(max) / Math.log(10.0));
    int digits_now;
    if (valueNow == 0) {
      digits_now = 0; //Both these numbers are really 1 short of the number of digits.
    } else {
      digits_now = (int)Math.floor(Math.log(valueNow) / Math.log(10.0));
    }
    for (int i = 0; i < max_digits - digits_now; i++) {
      buf.append('0');
    }
    buf.append(valueNow);
    return buf.toString();
  }
  
  
  public InputStream toInputStream (String pathname1, boolean testOnly_p)
  {
     InputStream stream1 = null;
     try {
       System.out.println("SqJoiner opening " + pathname1);    
       File testExistance = new File(pathname1);
       if (!testExistance.exists()) {
         throw new IOException("File DOES NOT EXIST - no read done.");
//       } else if (!testExistance.canWrite()) {
//         throw new IOException("File cannot be written to.");
       }
       if (testOnly_p)
         stream1 = null;
       else
         stream1 = new FileInputStream(pathname1);
     } catch (Throwable e) {
       System.out.println("SqJoiner FAILED TO OPEN " + pathname1);    
       stream1 = null;
       abort_ = true;
     } finally {
       return stream1;
     }
  }
  
  public OutputStream toOutputStream (String pathname1, boolean testOnly_p)
  {
     OutputStream stream1 = null;
     try {
  System.out.println("Opening output file " + pathname1);    
       File testExistance = new File(pathname1);
       if (testExistance.exists()) {
         throw new IOException("Already Exists or invalid - no write done.");
       }
       if (testOnly_p)
         stream1 = null;
       else
         stream1 = new FileOutputStream(pathname1);
     } catch (Throwable e) {
       messenger_.message("SqJoiner.toOutputStream " + 
                          (e.getMessage() == null ? e.toString() : e.getMessage()));
       stream1 = null;
     } finally {
       return stream1;
     }
  }

   
  private void beginOutput() throws Throwable
  {   
    
    for(int j = 0; j < col_count_; j++) {
      totalWidth_ += widths_[j];
    }

    System.out.println("Begin writing preparation.");
    FractalSpec outSpec = null;
    if (args_ != null) {
      // Set up the output spec. - This needs work.
      outSpec = new FractalSpec(args_);
      outSpec.pix_height = args_.pix_height * row_count_;
      outSpec.pix_width = totalWidth_;
      outSpec.panels = 1;
      outSpec.panel_x = 0;
      outSpec.panel_y = 0;
    
      System.out.println("SqJoiner.beginOutput outSpec.pix_height = " + outSpec.pix_height +
                         " row_count_ = " + row_count_ + " vLast_ = " + vLast_ +
                         " vFirst_ = " + vFirst_);
    }
    // Write out the images.
    outStream_ = toOutputStream(outputString_, false);
    // I think the 3rd arg below is pointless.  We are using color map for this.
    encoder_ = new GIFEncoder(totalWidth_, panel_height_ * row_count_, color_map_);
    encoder_.javaQuatWriteHeader(outStream_,
                                 (args_ != null ? 
                                   outSpec.toString(1, " Chunck Rectangular Composite") :
                                   null),
                                 false,
                                 transparentPixelIndex_);
  }
  
  
  private void attemptOutput() throws Throwable
  {   
    
    // Set up the output spec. - This needs work.
    System.out.println("Appending a row of panels.");
      
    // Write out the images.
    for(int ht = 0; ht < panel_height_; ht++) {
      if (abort_) {
        break;
      }
      for(int hz = 0; hz < col_count_; hz++) {
        if (abort_) {
          break;
        }
        int row_width = widths_[hz];
        int start = row_width * ht;
        // Last arg is exclusive.
        if (((byte[])pixels_[hz]).length < start + row_width - 1) {
          throw new IOException("SMALL COMPONENT ENCOUNTERED: " + 
                                " Y = " + (vFirst_ + vIndex_) +
                                " X = " + hz +
                                " ht = " + ht +
                                " panel_height_ = " + panel_height_ +
                                " pixels in panel = " + ((byte[])pixels_[hz]).length +
                                " start = " + start);
        }
        encoder_.javaQuatWritePixels((byte[])pixels_[hz], start, start + row_width);
      }
      if (ht % 10 == 0) {
        System.out.println("  Row " + ht + " written");
      }
    }
    if (abort_) {
      System.err.println("PANEL ROW WAS ABORTED.");
    } else {
      System.out.println("PANEL ROW COMPLETE.");
    }
  }

  private void finishOutput ()
  {
//  System.out.println("finishOutput entry.");
    try {
      if (abort_) {
        System.err.println("JOIN WAS ABORTED.  DELETE LAST FILE OUTPUT.");
      } else {
        encoder_.javaQuatWriteFooter();
        System.err.println("JOIN IS COMPLETE.");
      }
    } catch (Throwable e) {
      System.err.println("finishOutput error: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
    
      e.printStackTrace();
    }finally {
      if (outStream_ != null)
        try {
          outStream_.close(); // We didn't close it above.
        } catch (Throwable e) {}
    } 
  }      
}
