package JQJoiner;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.math.*;
import java.net.*;
import java.util.*;
import JQJoiner.GIFDecoderPkg.*;
import JQJoiner.GIFEncoderPkg.*;
import JQJoiner.TIFFEncoderPkg.*;

public class RectJoiner extends Canvas implements Runnable
{

  private String inputString_ = null;
  private String outputString_ = null;
  private Vector panelStrings_;
  int vFirst_;
  int vLast_;
  int hFirst_;
  int hLast_;
  
  private Messenger messenger_ = null;
  private FractalSpec args_ = null; // Stores the zeroth FractalSpec.
  private RGBxSizeSequence color_map_ = null;
  private int row_count_ = 0;
  private int col_count_ = 0;

  private int vIndex_;
  private int hIndex_;
  private int totalWidth_ = 0;
  private int row_width_ = 0; // Computer for each row if component images & checked 
                              // against totalWidth_.
  private int panel_height_ = 0;

  private boolean abort_ = false;
  private GIFEncoder gEncoder_ = null;
  private TIFFEncoder tEncoder_ = null;
  private OutputStream outStream_ = null;
  
  private Object [] pixels_;
  private int [] widths_;
  private int height_ = -1; // Values should all be the same.
  private boolean running_p_ = false;
  
  private boolean transparentPixelP_ = false;
  private int transparentPixelIndex_ = 0;
  
  private boolean tiffP_ = false;
  private int resolution_ = 72;
  private int tiffRowsPerStrip_ = 0;
  private int tiffCompression_ = TIFFEncoder.PACKBITS;
  private String copyRightNotice_ = null;
  
  
  RectJoiner (String inputString, int v1, int vn, int h1, int hn, String outputString)
    // Constructor.
  {
    inputString_ = inputString;
    vFirst_ = v1;
    vLast_ = vn;
    hFirst_ = h1;
    hLast_ = hn;    
    outputString_ = outputString;
    
    row_count_ = vLast_ - vFirst_ + 1;
    col_count_ = hLast_ - hFirst_ + 1;
    System.out.println("RectJoiner.constructor  row_count_ = " + row_count_ + " vLast_ = " +
                       vLast_ + " vFirst_ = " + vFirst_);

  }
  
  public void useTiff (boolean value)
  {
    tiffP_ = value;
  }
  
  public void setResolution (int res)
  {
    resolution_ = res;
  }
  
  public void setTiffRowsPerStrip (int value)
  {
    tiffRowsPerStrip_ = value;
  }
  
  public void setTiffCompression (int value)
  {
    tiffCompression_ = value;
  }
  
  public void setCopyRightNotice_ (String value)
  {
    copyRightNotice_ = value;
  }

  public boolean join (Messenger messenger)
    // Called just before run..
  {
    messenger_ = messenger;
    abort_ = false;
    // runIt();
    
    if (running_p_) {
      // This isn't going to happen, since we always make a new SqJoiner at present.
      messenger_.message("** A join is already running.  No action taken.");
      return false;
    }
    running_p_ = true;
    return true;
  }    
    
    
  public void run ()
    // Top level action method.
  {
    System.out.println(" ");
    System.out.println("Joining " +  row_count_ * col_count_ + " pictures.");    
    pixels_ = new Object [col_count_];  //Individual arrays added in loadGIF.
    widths_ = new int [col_count_];    
    
    PathVector rowPathnames = 
        new PathVector(vFirst_, vLast_, inputString_, "+");
    try {
      for(vIndex_ = 0; vIndex_ < row_count_; vIndex_++) {
        if (abort_) {
          break;
        }
        panelStrings_ = 
          new PathVector(hFirst_, hLast_, 
                         (String)rowPathnames.elementAt(vIndex_), "*");
        row_width_ = 0;
        for(hIndex_ = 0; hIndex_ < col_count_; hIndex_++) {
          abort_ = false;
//          messenger_.clearMessages();
          loadGIF(hIndex_); 
          if (abort_) {
            break;
          }
        }
        if (abort_) {
          break;
        }  else {
          messenger_.message(" -- Outputting row " + (vFirst_ + vIndex_));
          if (vIndex_ == 0) {
            if (tiffP_) {
              beginTiffOutput();// Opens output, makes encoder, computes totalWidth_ used below.
            } else { // GIF
              beginOutput();  // Sets spec, opens output, makes encoder, computes totalWidth_ used below.
            }
          }
          if (row_width_ != totalWidth_) {
            throw new IOException("Component row has wrong width.");
          }
          attemptOutput(); // Writes pixels - all formats.
        }
      }
    } catch (Throwable e) {
      messenger_.message("RectJoiner.join: problem: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
      System.err.println("RectJoiner.join: problem: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
      e.printStackTrace();
    }
    finishOutput(); // Tries to close file.
    running_p_ = false;
    System.gc();
  }
  
  
  public void abortJoin ()
  {
    abort_ = true;
  }
  
  public void loadGIF (int pixIndex)
  {
    GIFDecoder gifDecoder = null;
    InputStream stream1 = null;
    String argsString = null;
    String pathname = (String)panelStrings_.elementAt(pixIndex);
    System.out.println(" ");
    // Here we try to open the stream and extract the arguments only.
    try {
      stream1 = toInputStream(pathname, false); // false => create stream.
      if (stream1 == null)
        throw new IOException("NOT FOUND: " + pathname);

      gifDecoder = new GIFDecoder(stream1);
      argsString = gifDecoder.getApplicationData("JavaQuat", "020");
      args_ = null;
      if (argsString != null) {
        StringReader argsStream = new StringReader(argsString);
        JQArgReader reader = new JQArgReader();
        // parent.exitImageSaver gets passed argsReadFromFile_ on exitButton.
        args_ = new FractalSpec(reader.readArgFile(argsStream, -1, messenger_));
      }
      if (args_ == null) {
        System.out.println("RectJoiner.loadGIF: Empty spec returned by loadGIF");
      }

      if (pixIndex == 0) {
        panel_height_ = gifDecoder.getHeight();
      } else if (panel_height_ != gifDecoder.getHeight()) {
        throw new IOException("Component picture has wrong height: " + gifDecoder.getHeight() + 
                              ".  Should be " + panel_height_ + " like 1st row component.");
      }
      widths_[pixIndex] = gifDecoder.getWidth();
      row_width_ += widths_[pixIndex];
      System.out.println("  RectJoiner:  which is image " +  pixIndex +
                         " of height " + gifDecoder.getHeight() + " and width " + widths_[pixIndex]);    
      
      pixels_[pixIndex] = gifDecoder.pixelsAsIndices();
      if (pixels_[pixIndex] == null) {
        System.out.println("RectJoiner.loadGIF: Empty pixels_[" +
                           pixIndex + "]");
        abort_ = true;
        throw new IOException("Empty pixels_ in loadGIF");
      }
    } catch (Throwable e) {
      messenger_.message("RectJoiner.loadGIF" +
                         (e.getMessage() == null ? e.toString() : e.getMessage()));
      System.out.println("RectJoiner.loadGIF: catch " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()));
      e.printStackTrace();
      abort_ = true;
    } finally {
      if (stream1 != null)
        try {
          stream1.close(); // In case we didn't close it above.
        } catch (Throwable e) {}
    }  
    
    if (pixIndex == 0 && vIndex_ == 0) {
      System.out.println("  Preparing color map.");    
      // Set up the byte and int versions of the colormap.
      color_map_ = gifDecoder.globalRGBTable();
      if (color_map_ == null) {
        System.err.println("RectJoiner.loadGIF: Failed to set color_map_.");
        abort_ = true;
      }
      transparentPixelP_ = gifDecoder.transparentPixelP();
      if (transparentPixelP_) {
        transparentPixelIndex_ = gifDecoder.transparentPixelIndex();
      } else {
        transparentPixelIndex_ = -1;
      }
    }
  }
    
  public InputStream toInputStream (String pathname1, boolean testOnly_p)
  {
     InputStream stream1 = null;
     try {
       System.out.println("RectJoiner opening " + pathname1);    
       File testExistance = new File(pathname1);
       if (!testExistance.exists()) {
         throw new IOException("File DOES NOT EXIST - no read done.");
//       } else if (!testExistance.canWrite()) {
//         throw new IOException("File cannot be written to.");
       }
       if (testOnly_p)
         stream1 = null;
       else {
         stream1 = new FileInputStream(pathname1);
         stream1 = new BufferedInputStream(stream1);
       } 
     } catch (Throwable e) {
       System.out.println("RectJoiner FAILED TO OPEN " + pathname1);    
       stream1 = null;
       abort_ = true;
     } finally {
       return stream1;
     }
  }
  
  public OutputStream toOutputStream (String pathname1, boolean testOnly_p)
  {
     OutputStream stream1 = null;
     try {
  System.out.println("Opening output file " + pathname1);    
       File testExistance = new File(pathname1);
       if (testExistance.exists()) {
         throw new IOException("Already Exists or invalid - no write done.");
       }
       if (testOnly_p)
         stream1 = null;
       else {
         stream1 = new FileOutputStream(pathname1);
         stream1 = new BufferedOutputStream(stream1);
       }
     } catch (Throwable e) {
       messenger_.message("RectJoiner.toOutputStream " + 
                          (e.getMessage() == null ? e.toString() : e.getMessage()));
       stream1 = null;
     } finally {
       return stream1;
     }
  }

   
  private void beginOutput() throws Throwable
  {   
    
    for(int j = 0; j < col_count_; j++) {
      totalWidth_ += widths_[j];
    }

    System.out.println("Begin writing preparation.");
    FractalSpec outSpec = null;
    if (args_ != null) {
      // Set up the output spec. - This needs work.
      outSpec = new FractalSpec(args_);
      outSpec.pix_height = args_.pix_height * row_count_;
      outSpec.pix_width = totalWidth_;
      outSpec.panels = 1;
      outSpec.panel_x = 0;
      outSpec.panel_y = 0;
    
      System.out.println("RectJoiner.beginOutput outSpec.pix_height = " + outSpec.pix_height +
                         " row_count_ = " + row_count_ + " vLast_ = " + vLast_ +
                         " vFirst_ = " + vFirst_);
    }
    // Open stream and prepare encoder
    outStream_ = toOutputStream(outputString_, false);
    // I think the 3rd arg below is pointless.  We are using color map for this.
    gEncoder_ = new GIFEncoder(totalWidth_, panel_height_ * row_count_, color_map_);
    gEncoder_.javaQuatWriteHeader(outStream_,
                                 (args_ != null ? 
                                   outSpec.toString(1, " Rectangular Composite") :
                                   null),
                                 false,
                                 transparentPixelIndex_);
  }
  
  private void beginTiffOutput () throws Throwable
    //beginOutput for tiff
  {
    for(int j = 0; j < col_count_; j++) {
      totalWidth_ += widths_[j];
    }

    System.out.println("Begin TIFF writing preparation.");
    final int bitsPerPixel = 8;
    final int indexedColors = 256;
    byte [] reds = new byte [indexedColors];
    byte [] greens = new byte [indexedColors];
    byte [] blues = new byte [indexedColors];
    
    //System.out.println("Begin RectJoiner.beginTiffOutput. color_map_=" + color_map_);
    int height = panel_height_ * row_count_;
    int width = totalWidth_;
    int i = 0;
    for (i = 0; i < color_map_.byteTable.length / 3; i++) {
      reds[i] = color_map_.byteTable[i*3];
      greens[i] = color_map_.byteTable[i*3 + 1];
      blues[i] = color_map_.byteTable[i*3 + 2];
    }
    for (; i < indexedColors; i++) {
      reds[i] = 0;
      greens[i] = 0;
      blues[i] = 0;
    }

    //System.out.println("Begin RectJoiner.beginTiffOutput. pt3");
    // Open stream and prepare encoder
    outStream_ = toOutputStream(outputString_, false);

    //System.out.println("Begin RectJoiner.beginTiffOutput. pt4");
    IndexColorModel colorModel = new IndexColorModel(bitsPerPixel, indexedColors, reds, greens, blues);
    //System.out.println("Begin RectJoiner.beginTiffOutput. pt5");
    tEncoder_ = new TIFFEncoder(width, height, colorModel);
    tEncoder_.setVerbose(true);
    tEncoder_.setResolution(resolution_);
    if (copyRightNotice_ != null) {
      tEncoder_.setCopyrightNotice(copyRightNotice_);
    }
    //System.out.println("Begin RectJoiner.beginTiffOutput. pt6");
    tEncoder_.setCompression(tiffCompression_);
    if (tiffRowsPerStrip_ > 0) {
      tEncoder_.setStripParameters(tiffRowsPerStrip_);
    } else {
      tEncoder_.setStripParameters(panel_height_);
    }
    //System.out.println("Begin RectJoiner.beginTiffOutput. pt7");
  }
  
  
  private void attemptOutput() throws Throwable
  {   
    
    // Set up the output spec. - This needs work.
    System.out.println("Appending a row of panels.");
    
    // Write out the images.
    for(int ht = 0; ht < panel_height_; ht++) {
      if (abort_) {
        break;
      }
      for(int hz = 0; hz < col_count_; hz++) {
        if (abort_) {
          break;
        }
        int row_width = widths_[hz];
        int start = row_width * ht;
        
        /*if (ht==0) {
          System.out.println("RectJoiner.attemptOutput hz=" + hz +
                             " 0th pixel is " + ((byte[])pixels_[hz])[0]);
        }*/
        
        // Last arg is exclusive.
        if (((byte[])pixels_[hz]).length < start + row_width - 1) {
          throw new IOException("SMALL COMPONENT ENCOUNTERED: " + 
                                " Y = " + (vFirst_ + vIndex_) +
                                " X = " + hz +
                                " ht = " + ht +
                                " panel_height_ = " + panel_height_ +
                                " pixels in panel = " + ((byte[])pixels_[hz]).length +
                                " start = " + start);
        }
        if (tiffP_) {
          tEncoder_.write(outStream_, (byte[])pixels_[hz], start, row_width);
        } else {
          gEncoder_.javaQuatWritePixels((byte[])pixels_[hz], start, start + row_width);
        }
      }
      if (ht > 0 && ht % (tiffP_ ? 100 :10) == 0) {
        System.out.println("  Row " + ht + (tiffP_ ? " accumulated" : " written"));
      }
    }
    if (abort_) {
      System.err.println("PANEL ROW WAS ABORTED.");
    } else {
      System.out.println("PANEL ROW COMPLETE.");
    }
  }

  private void finishOutput ()
  {
    if (tiffP_) {
      System.err.println("JOIN IS COMPLETE.");
      try {
        outStream_.close();
      } catch (Throwable e) {}
      return;
    }
//  System.out.println("finishOutput entry.");
    try {
      if (abort_) {
        System.err.println("JOIN WAS ABORTED.  DELETE LAST FILE OUTPUT.");
      } else {
        gEncoder_.javaQuatWriteFooter();
        System.err.println("JOIN IS COMPLETE.");
      }
    } catch (Throwable e) {
      System.err.println("finishOutput error: " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
    
      e.printStackTrace();
    }finally {
      if (outStream_ != null)
        try {
          outStream_.close(); // We didn't close it above.
        } catch (Throwable e) {}
    } 
  }      
}
