package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class Loop25 extends JQLooper
{
  public String formulaString () {return "Z(n+1) = Z(n)^5 + C2*Z(0) + T";}
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return " Mandelbrot-like fractal using 5th power in the complex plane";}
  public String q1String () {return "Z(0):r/i/-/-";}
  public String q2String () {return "C:r/i/-/-";}
  public String q3String () {return "T:r/i/-/-";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    FastQuaternion add = coeff.times(point).plus(trans);
    double z_r = point.r;
    double z_i = point.i;
    double add_r = add.r;
    double add_i = add.i;
    double real;
    double zrsq = z_r * z_r;
    double zisq = z_i * z_i;
    double magnitude;
    for (int i = 1; i < depth; i++) {
      real = zrsq * zrsq * z_r - 10 * zrsq * zisq * z_r + 5 * zisq * zisq * z_r + add_r;
      z_i =  5 * zrsq * zrsq * z_i - 10 * zrsq * zisq * z_i + zisq * zisq * z_i + add_i;
      z_r = real;
      zrsq = z_r * z_r;
      zisq = z_i * z_i;
      if (zrsq + zisq > 4)
        return i;
    }
    return depth;
  }
}

