package JavaQuat20.FractalLoops;

import JavaQuat20.*;

public class QuatPower5 extends JQLooper
  // Quaternion space.
{
  public FastQuaternion q1Init () {return new FastQuaternion(0,0,0,0);}
  public FastQuaternion q2Init () {return new FastQuaternion(1,0,0,0);}
  public FastQuaternion q3Init () {return new FastQuaternion(0,0,0,0);}
  public String docString () {return "\nQuaternion multiplication table:\n" +
  "  __________________________\n" +
  "   * ||  r |  i |  j |  k |\n" +
  " ==========================\n" +
  " | r ||  r |  i |  j |  k |\n" +
  " __________________________\n" +
  " | i ||  i | -r |  k | -j |\n" +
  " __________________________\n" +
  " | j ||  j | -k | -r |  i |\n" +
  " __________________________\n" +
  " | k ||  k |  j | -i | -r | \n" +
  " __________________________";}
  public String formulaString () {return "Z(n+1) = Z(n)^5 + C*Z(0) + T - using quaternion space";}
  public String q1String () {return "Z(0)";}
  public String q2String () {return "C";}
  public String q3String () {return "T";}

  public int loop (FastQuaternion point, FastQuaternion coeff, FastQuaternion trans, int depth)
   // Returns value in range [1, depth] inclusive.
  {
    double add_r = point.r * coeff.r - point.i * coeff.i - point.j * coeff.j - point.k * coeff.k + trans.r;
    double add_i = point.r * coeff.i + point.i * coeff.r + point.j * coeff.k - point.k * coeff.j + trans.i;
    double add_j = point.r * coeff.j - point.i * coeff.k + point.j * coeff.r + point.k * coeff.i + trans.j;
    double add_k = point.r * coeff.k + point.i * coeff.j - point.j * coeff.i + point.k * coeff.r + trans.k;
    double r = point.r;
    double i = point.i;
    double j = point.j;
    double k = point.k;
    double rsq;
    double isq;
    double jsq;
    double ksq;
    double new4r;
    double new4i;
    double new4j;
    double new4k;
    double r5;
    double i5;
    double j5;
    double quatSq_r;
    for (int z = 1; z < depth; z++) {
      rsq = r * r;
      isq = i * i;
      jsq = j * j;
      ksq = k * k;
      if (rsq + isq + jsq + ksq > 2)
        return z;
      quatSq_r = rsq - isq - jsq - ksq;
      new4r = quatSq_r * quatSq_r - 4 * rsq * (isq + jsq + ksq);
      new4i = 4 * r * i * quatSq_r;
      new4j = 4 * r * j * quatSq_r;
      new4k = 4 * r * k * quatSq_r;
      r5 = r * new4r - i * new4i - j * new4j - k * new4k + add_r;
      i5 = r * new4i + i * new4r + j * new4k - k * new4j + add_i;
      j5 = r * new4j - i * new4k + j * new4r + k * new4i + add_j;
      k =  r * new4k + i * new4j - j * new4i + k * new4r + add_k;
      j = j5;
      i = i5;
      r = r5;
    }
    return depth;
  }
  
  /*
  private FastQuaternion unique_times_equal (FastQuaternion self, FastQuaternion point)
  {
    double rr, ii, jj, kk;
    rr = self.r * point.r - self.i * point.i - self.j * point.j + self.k * point.k;
    ii = self.r * point.i + self.i * point.r - self.j * point.k - self.k * point.j;
    jj = self.r * point.j - self.i * point.k + self.j * point.r - self.k * point.i;
    kk = self.r * point.k - self.i * point.j - self.j * point.i + self.k * point.r;
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
  
    
   private FastQuaternion unique_sq_self (FastQuaternion self)
  {
    double rr, ii, jj, kk;
    rr = self.r * self.r - self.i * self.i - self.j * self.j + self.k * self.k;
    ii = 2 * (self.r * self.i - self.j * self.k);
    jj = 2 * (self.r * self.j - self.k * self.i);
    kk = 2 * (self.r * self.k - self.i * self.j);
    self.r = rr;
    self.i = ii;
    self.j = jj;
    self.k = kk;
    return self;
  }
   */
 
  
  /*  Times table
  __________________________
  Quat*||  r |  i |  j |  k |
  ==========================
  | r ||  r |  i |  j |  k |
  __________________________
  | i ||  i | -r |  k | -j |
  __________________________
  | j ||  j | -k | -r |  i |
  __________________________
  | k ||  k |  j | -i | -r | 
  __________________________
  */
  
}

